///////////////////////////////////////////////////////////////////////////
// Canon EOS Shutter-Emulation on Arduino (Nano / Pro Mini, / ...)
//
//       25.1.2017 - 31.01.2025                       Erwin Lotter
//

// EOS emulation core for single coil use, to be included from ShutterEmuEOS.ino

  //
  // Ablaufschema:
  // ------------------------------------
  // Vor einer Auslsung oder bei lnger geffnetem Shutter
  //  befindet sich der Prozessor im Standby, aus der ihn eine Flanke von Coil 1 oder
  //  ggf. Coil 2 weckt und in triggerEvent einen Code eintrgt.
  //
  // Der Trigger weckt auch den Prozessor, so dass loop() wieder durchlaufen wird,
  //  wo dann die zu triggerEvent passende Aktionsliste gestartet wird.
  //
  // Bis zum Ende der aktuellen Pulsgruppe wird loop() durchlaufen und geprft,
  //  ob die in actionPtr[stateIndex].Time() vermerkte Zeit erreicht ist. 
  //  Wenn ja, wird die hinterlegte Aktion ausgefhrt und stateIndex erhht.
  //  LONGWAIT_VAL und HALT_VAL Eintrge halten die Ausfhrung bis zum nchsten
  //  Trigger an, STOP_VAL beendet die laufende Aktionsliste.
  //
  // Ablauf eines Ereignisses mit Coil 1 (6D, 80D: Coil 2) als Trigger:
  // ------------------------------------------------------------
  // 1.) Zuerst wird geprft, ob Coil 1 eine nderung erfahren hat (triggerEvent wurde im Int-Handler gesetzt).
  //  a)  1 -> 0   stateIndex wird auf 0 gesetzt, timingZero erhlt die aktuelle Zeit.
  //               (Viele Modelle schalten regelmig, andere nur bei Half-Shoot, die Coil-Pwr aktiv.
  //		    Dadurch erzeugen reale Coils Signale, die aber mit einem Emulator nicht auftreten, weil die Coils
  //                gegen einen konstanten Pullup arbeitet, die emulierte Coil-Pwr also immer anliegt!).
  //  b)  0 -> 1   Zweiter Teil (Schlieen) beginnt, stateIndex wird bis hinter einen eventuellen LONGWAIT
  //               inkrementiert. Noch anstehende Aktionen werden dabei ausgefhrt.
  //               Die positive Coil-Flanke setzt auch die Zeitbasis fr den Rest der Signale neu.
  //
  // Mgliche Folgen mit Coil 1 allein:
  //  c1_L_Start - c1_H_FinishShoot
  //  c1_L_Start - c1_H_EnterLV
  //  c1_L_StartInLV - c1_H_FinishLVShoot  +  c1_L_StartInLV - c1_H_EnterLV
  //  c1_L_StartInLV - c1_H_LeaveLV
  //
  // Mgliche Folgen mit Coil 2 (pos. Flanke) allein:
  //  c1_L_Start - c2_H_FinishShoot (mit LONGWAIT bis c1_H)
  //  c1_L_Start - c2_H_EnterLV
  //  c1_L_StartInLV - c2_H_FinishLVShoot (mit LONGWAIT bis c1_H) +  c1_L_StartInLV - c2_H_EnterLV
  //  c1_L_StartInLV - c2_H_LeaveLV  (nicht fr 600D: da bleibt C2 inaktiv
  //
  // Der Vorteil von c2_H liegt darin, dass es unabhngig von der Belichtungszeit zu
  //  einem festen Zeitpunkt kommt. Wahlweise knnen die c2_H Timingtabellen, die auf
  //  c1_H warten mssen, ein LONGWAIT enthalten, oder das c1_H Event whlt eine neue
  //	Tabelle.
  //
  if (triggerEvent == evCoil1Start) {        // Start-Flanke (Coil 1 - oder, wenn 6D, Coil 2 - aktiviert):
    InitEvent();
    stateIndex = 0;
    //
    // Auerhalb des LV kann:
    //  a) eine normale Auslsung vorliegen, die mit c1_L_Start behandelt wird, oder
    //  b) ein Start des LV, fr den c1_L_Start die ROT-Signale zu weit ausfhrt.
    //  c1_L_Start stoppt nach 48 ms und wartet auf Coil 2 --> Hi. Der C2_H Zeitpunkt
    //   lsst zwischen a) und b) unterscheiden.
    //
    // Im LV war die letzte Aktion entweder
    //  - coilX_FinishLVShoot, dem wieder ein cX_H_EnterLV folgen muss, oder
    //  - cX_H_EnterLV. Danach kommt entweder
    //  	a) LV-Auslsen oder 
    //  	b) LV-Ende. Dafr ist c1_L_Start ungeeignet, deshalb c1_L_StartInLV verwenden:
    //
    if (actionPtr == c1_H_EnterLV	// c1_H_EnterLV kann auch im Coil2-Modus vorkommen, wenn die nicht angeschlossen ist
	// Achtung, fr Coil 1 msste hier eigentlich kommen:				??600D
	//|| actionPtr == c1_H_FinishLVShoot
	// aber	das klappt fr die 600D nicht (Err20)!!
	// Ohne diese Prfung wird nach LV-Shoot das jetzt folgende Re-Enter-LV mit c1_L_Start 
	//  statt mit c1_L_StartInLV (erfolgreich) eingeleitet. Dadurch wird dann bei C1_H das
	//  erste c1_H_EnterLV anstatt dem zweiten (dem Re-EnterLV) ausgewhlt. Die sind eh gleich.
	// Der Fehler tritt auch auf, wenn c1_L_StartInLV gleich ist wie c1_L_Start. Er besteht
	//  vmtl. darin, dass nach c1_H_FinishLVShoot das C1_H Timing ncht richtig erkannt wird.
	// besser so Cam-spezifisch def.:
	//#define IS_IN_LV 		(actionPtr == cX_H_EnterLV  ||  actionPtr == c1_H_FinishLVShoot)
	// 
    ) {
      actionPtr = c1_L_StartInLV;
      //PRINTLN("c1_L_StartInLV");
    }
    else {
      actionPtr = c1_L_Start;
      //PRINTLN("c1_L_Start");
    }
    NotifyIntervalTimer();
  } // if (triggerEvent == evCoil1Start) -----------------------------------------------
  
  else if (triggerEvent == evCoil1Stop) { // Coil 1 wurde inaktiv (C1_H):
    //
    // Falls eine Belichtung stattfand, sollte aktuell ein LONGWAIT 
    //  laufen.
    //
    {
      unsigned long dt = triggerTime - timingZero;		// Coil-1 Pulsdauer
      if (actionPtr == c1_L_Start) {		// (noch) nicht im LV?
	if (IsInWindow_EnterLV(dt)) {
	  actionPtr = c1_H_EnterLV;		// dt passt zu LV-Start oder Restart -- FALSCH bei best. Belichtungszeiten (600D: < 1/60s!!!)
	  //PRINTLN("c1_H_EnterLV");
	}
	else {
	  actionPtr = c1_H_FinishShoot;	// zu kurz/lang fr EnterLV: regulres (nicht zu kurzes!) Auslsen
	  //PRINTLN("c1_H_FinishShoot");
	}
      }
      else { // im LV:
	      // a) Shoot im LV endet
	      // b) Re-Enter LV endet (folgt immer nach a)
	      // c) Leave LV endet: Kann nur nach EnterLV kommen, dt ist sehr kurz
	      //
	if (IsInWindow_LeaveLV(dt)) {		// LeaveLV hat krzesten Puls, also zuverlssig zu erkennen!
	  //PRINTLN("c1_H_LeaveLV");
	  actionPtr = c1_H_LeaveLV;
	}
	else if (IsInWindow_ReEnterLV(dt)
	  #ifdef EOS600D
		 &&  actionPtr != c1_L_StartInLV  // fr 600D zustzlich prfen: kurze Zeiten im LV liegen unter HiLIMIT_C1_LEAVE_LV
	  #endif	  
	) {
	  //PRINTLN("c1_H_(Re)EnterLV");
	  actionPtr = c1_H_EnterLV;
	}
	else {
	  // zu lange fr EnterLV: regulres (nicht zu kurzes!) Auslsen
	  // geht fr 600D schief fr kurze Belichtungen. Sicher zu unterscheiden, wenn vorheriger actionPtr gespeichert wird: Der war dann nicht
	  //  c1_H_FinishLVShoot, weil danach immer c1_H_EnterLV folgt
	  //PRINTLN("c1_H_FinishLVShoot");
	  actionPtr = c1_H_FinishLVShoot;
	}
      }
      stateIndex = 0;
    }
    InitEvent();
  } // if (triggerEvent == evCoil1Stop) -----------------------------------------------------
