/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"

#if defined(SPIFLASH_CS)	    // indicates SPI flash module available

#include "FlashUpdate.h"
 
BootFlash bootFlash(SPIFLASH_CS);

//#define DEBUG_FLASH	  // Achtung: Fördert dynamic memory shortage!

// Zugriff auf den Flashspeicher init.
void IntervallTimer::setupSPIFlash()
{
  bootFlash.begin();
  #if defined(DEBUG_FLASH)
  Serial.begin(57600);
  #endif
}


//
// Upload eines Sketches ins externe Flash initialisieren.
// Dazu auch die ersten 32K des Flash löschen.
// 
// Rückgabe: False wenn Fehler beim Löschen
//
bool BootFlash::InitCodeUpload()		      
{
  digitCnt = -1; 
  return Erase32K(0);
}


//
// 32K des Flash löschen ab addr.
// 
// Rückgabe: False wenn Fehler beim Löschen
//
bool BootFlash::Erase32K(uint32_t addr)		      
{
  if (!eraseBlock32K(addr)) {
    for (int i = 0; i < 8; i++) {  // sector erase time is 400ms
      if (!eraseSector(addr + (i << 12))) {	    // 4K
	return false;
      }
    }
  }
  return true;
}


//
// Bereich innerhalb des Flash kopieren (ohne vorheriges Löschen).
// size muss Vielfaches von 128 Byte sein!
// 
// 32KB dauern < 1/2s (Einzelbytetransfer braucht 13s).
//
bool BootFlash::CopyFlashToFlash(uint32_t from, uint32_t to, uint16_t size)
{
  const int blocksizepwr = 7;	// 128 bytes per call
  const int blocksize = (1<<blocksizepwr);
  uint8_t bf[blocksize];
  uint16_t i = size >> blocksizepwr;   // Anzahl von 128 Byte Stücken, abgerundet
  while (   i > 0
	 && readByteArray(from, bf, blocksize, false)
	 && writeByteArray(to, bf, blocksize, true)
  ) {
    from += blocksize;
    to   += blocksize;
    i--;
  }
  return i==0;
/* // Einzelbytetransfer
  uint16_t i = size;
  while (   i > 0
	 && writeByte(to, readByte(from, false), true)
  ) {
    from ++;
    to   ++;
    i--;
  }
  return i==0;
//*/
}


//
// Flashinhalt anzeigen als Hex-Dump
// pmem wählt die Seite aus:
//  0 = ab Anfang des internen Flash
//  1 = ab Anfang des externen Flash (Upload-Puffer)
//  2 = Fallback-Puffer im externen Flash (ab FALLBACK_FLASHPAGE)
//  3 = Protokoll-Flash (s. PROTOCOL_FLASHPAGE)
//
void IntervallTimer::PrintFlash(int pmem)
{
  uint8_t page = 0;
  if (pmem == 3) {
    page = PROTOCOL_FLASHPAGE >> 16;
  }
  for (uint16_t x = 0 ; x < 0x0100 ; x++) {
    if ((x & 0xf) == 0) {
      bluePrintln();
      if (pmem > 0) {
	blueSerial.write('f');
	bluePrintHexDigit(page); 
      }
      bluePrintHex(x);
      bluePrint_P(PSTR(":"));
    }
    if ((x & 0x3) == 0) {     // am Anfang und alle 4 Bytes
      bluePrint_P(PSTR(" ")); //  noch ein Space
    }
    if (pmem==0) {
      bluePrintHex(pgm_read_byte(x));
    }
    else if (pmem==3) {
      bluePrintHex(bootFlash.readByte(PROTOCOL_FLASHPAGE+x));
    }
    else if (pmem==2) {
      bluePrintHex(bootFlash.readByte(FALLBACK_FLASHPAGE+x));
    }
    else {
      bluePrintHex(bootFlash.readByte(x));
    }
    bluePrint_P(PSTR(" "));   // Abstand zum nächsten Byte
  }
  blueSerial.println();
}


//
// Upload vervollständigen und Reset zur Installation als
//  Update (durch Urboot) auslösen.
//
bool IntervallTimer::SendUploadToBootloader()
{
  if (!bootFlash.CompleteUpload()) {
    return false;
  }
  watchdogMode = 1;		  // watchdog installiert fallback, falls Update hängt
  SaveSettingsToEEPROM();
  //EnterNewLine(PSTR("Reset wird ausgeführt .."));
  boardReset();
  return true;
}


#if defined(FALLBACK_SUPPORT)
//
// Upload vervollständigen und in der zweiten Hälfte der
//  ersten Flash-Page ablegen
//
bool IntervallTimer::SendUploadToFallback()
{
  watchdogDisable();	  // falls Kopie lange dauert (wird auch nicht gebraucht, wenn Upload geklappt hat)
  if (   bootFlash.Erase32K(FALLBACK_FLASHPAGE)
      && bootFlash.CopyFlashToFlash(0, FALLBACK_FLASHPAGE, FALLBACK_SIZE)
      && bootFlash.WriteFirstWord(FALLBACK_FLASHPAGE)
  ) {
    EnterNewLine(PSTR("Fallback eingerichtet."));
    return true;
  }
  bootFlash.SendError();
  return false;
}

//
// Fallback-Version an den Flash-Anfang kopieren.
//
bool BootFlash::CopyFallbackTo0x000()
{
  return Erase32K(0)
    /*
      && CopyFlashToFlash(FALLBACK_FLASHPAGE, 0, FALLBACK_SIZE-128)
    /*/
      && CopyFlashToFlash(FALLBACK_FLASHPAGE+2, 2, FALLBACK_SIZE-2)
      && ReadFirstWord(FALLBACK_FLASHPAGE)
      && WriteFirstWord(0x0000)
    //*/
    ;
}

//
// Fallback-Version an den Flash-Anfang kopieren und Reset
//  zur Installation als Update (durch Urboot) auslösen.
// Prüft nicht, ob tatsächlich eine Fallback-Version vorhanden
//  ist, weil Urboot dann eh nichts tun würde.
//
bool IntervallTimer::SendFallbackToBootloader()
{
  watchdogDisable();	  // falls Kopie lange dauert (wird auch nicht gebraucht, wenn Upload geklappt hat)
  // Im externen Flash:
  // 1.) 0000-7FFF löschen 
  // 2.) 8002-FFFF nach 0002 kopieren
  //     (Achtung: CopyFlashToFlash() kopiert nur komplette 
  //	  128-Byte Blöcke. Macht aber nichts, weil die letzten 
  //	  128 Byte sowieso immer für den Bootloader reserviert sind.)
  // 3.) 8000-8001 holen und
  // 4.) nach 0000-0001 schreiben (zur Sicherheit erst zum Schluss).
  //
  EnterNewLine(PSTR("Fallback aktivieren..."));
  if (bootFlash.CopyFallbackTo0x000()) {
    boardReset();
    return true;
  }
  bootFlash.SendError();
  return false;
}
#endif


void BootFlash::SendError() 
{
  uploadState = 0;
  bluePrint_P(PSTR("\nFehler beim Flash-Zugriff!"));
}


// Neues Zeichen aus der .hex Datei aufnehmen und ggf. verarbeiten.
// (s. https://en.wikipedia.org/wiki/Intel_HEX)
// 
// Rückgabe: 1, wenn fertig, 0 wenn weitere Daten erwartet, < 0 bei Fehlern:
//  -1 : Record zu groß
//  -2 : Checksum
//  -3 : Record Typ unbekannt
//  -4 : Flash Write Error
//  -5 : Ungültiges Zeichen
//
int8_t BootFlash::AddToUpload(int8_t c)
{
  if (digitCnt < 0) {
    if (c == ':') {	      // neuer Record
      digitCnt = 0;
      return 0;
    }
    if (isspace(c)) {	      // cr, lf vor Records ignorieren
      return 0;
    }
    return -5;		      // andere Zeichen vor Records nicht erlauben
  }
  const char* hexdigs = "0123456789ABCDEF";
  uint8_t val = (uint8_t)(strchr(hexdigs, c) - hexdigs);
  if (val < 16) {	      // Hex-Ziffer?
    uint8_t* p = &lineBuffer[digitCnt >> 1];
    digitCnt++;		      // gelesene Digits (= 2 x Bytes)
    if (digitCnt & 1) {	      // ungerade?
      *p = val << 4;	      //  1. Digit eintragen
      if (digitCnt == 1) {    //  byte count noch nicht da?
	return 0;	      //   dann noch nichts zu tun
      }
    }
    else {		      // gerade Zahl von Digits:
      val += *p;	      //  Digits zu Byte zusammenfassen
      *p = val;
      if (digitCnt == 2) {    //  byte count gerade erhalten? (in lineBuffer[0])
	limit = (1 + 2 + 1 + GetByteCount() + 1) << 1;	// Byte count + Address + Record type + Data + Checksum
	if (limit > (ELEMENTS(lineBuffer) << 1)) {
	  return -1;	      // Record zu groß!
	}
	checksum = val;
	/*  führt zu Fehler beim Upload weil zu langsam:   !!!!!!!!!!
	const int pcnt = 1024 / 16;   // 16-Byte Records pro KB
	static int np = pcnt;
	if (np == 0) {
	  //blueSerial.write('.');      // dauert zu lange
	  blueSerial.writeThrough('.'); // auch nicht besser
	  np = pcnt;
	}
	np--;
	//*/
	return 0;
      }
      checksum += val;	      // Summe über alles (incl. Checksum)
    }
    if (digitCnt == limit) {  // Record vollständig:
      if (checksum != 0) {    // (Checksum ist 1 - byte_summe !)
	return -2;
      }
      switch (GetRecordType()) {  // Typ prüfen (s. https://en.wikipedia.org/wiki/Intel_HEX)
	default:		  // alle anderen Typen:
	  return -3;		  //  Fehler, sollte nicht vorkommen

	case 1:			  // fertig:
	  uploadState = 1;	  //  zeigt an, dass Upload erfolgreich war
	  return 1;		  //  fertig (falls die Datei noch weiter geht, wird sie normaler Input!)

	case 0: {		  // Code-Record:
	  uint8_t* pd = GetData();
	  uint8_t size = GetByteCount();
	  uint16_t addr = GetAddress();
	  // Write to flash:
	  if (addr == 0) {
	    firstBytes[0] = *pd++;     // erstes und zweites Byte noch nicht, sonst wird beim nächsten Reset
	    firstBytes[1] = *pd++;     // das Update durchgeführt, auch wenn etwas schiefgeht
	    size -= 2;
	    addr = 2;
	  }
	  /* Nur für Vektor-Bootloader:
	  else if (addr == 0x0060) {
	    // Byte 0x0066 patchen: Wird sonst von avrdude erledigt
	    // Sehr unsicher!!
	    pd[6] -= 0xFC - 0xD4;     // 0xFC -> 0xD4, manchmal auch 0xFD -> 0xD5
	  }
	  */
	  if (!writeByteArray(addr, pd, size, true)) {
	    return -4;
	  }
	}
	case 3:		      // Start Segment Address (kommt in Urboot vor). The execution should start at this address:
			      //  the first two data bytes are the code segment, the latter two are the instruction pointer value.
			      //  Record ignorieren
	case 4:		      // 32-Bit Adresse: Wird zur Kennzeichnung von ...ino.with_bootloader.hex verwendet.
	  break;	      //  Record ignorieren
      }
      digitCnt = -1;	      //  neuen Record beginnen
      return 0;		      //  weiterlesen
    }
    return 0;		      // Record geht noch weiter
  }
  return -5;		      // andere Zeichen sind innerhalb eines Records nicht erlaubt
}

#endif

