///////////////////////////////////////////////////////////////////////////
// Canon EOS Shutter-Emulation on Arduino ffor EOSM models
//
// For details see ShutterEmuEOS.ino
//
//       25.1.2017 - 3.08.2024                       Erwin Lotter

#if defined EOS_MXXX	// switch between EOSM and EOS (code see main file ShutterEmuEOS.ino)


// Ablaufschema
// -------------
// Außerhalb der Pulsgruppen, die Events zugeordnet werden, befindet sich
//  der Prozessor im Standby, aus dem ihn eine Flanke von COIL1 oder MOTOR
//  per Interrupt weckt.
//
// Bis zum Ende der aktuellen Pulsgruppe wird loop() durchlaufen und geprüft,
//  ob die in actionPtr->Time() vermerkte Zeit erreicht ist. 
//  Wenn ja, wird die hinterlegte Aktion ausgeführt und actionPtr erhöht.
//
void loop() {
  //
  // Ablauf eines Ereignisses:
  // --------------------------
  // In den Interrupt-Handlern wird in triggerEvent hinterlegt, welches Ereignis auftrat. 
  // Passend dazu wird hier in actionPtr ein Zeiger auf die Signalsequenz hinterlegt.
  // Anschließend wird diese Sequenz abgearbeitet und abschließend der Sleep-Mode aktiviert, aus
  //  dem der Prozessor erst durch den nächsten Interrupt wieder geweckt wird.
  //
  if (triggerEvent != evNone) {        // neuer Interrupt?
    timingZero = micros();             //    Zeitpunkt festhalten
    if (triggerEvent == evCoil1Start) { // Start-Flanke Coil:
      actionPtr = coilStartAction;
      //PRINTLN("c1_L");
    }
    else if (triggerEvent == evCoil1Stop) {        // Stopp-Flanke (Coil inaktiv):
#ifdef EOS_M3
      // Für eine eventuelle Bulb-Auslösung wurde der Motor-Interupt aktiviert.
      // Falls er nicht ausgelöst wurde, muss er noch deaktiviert werden:
      // 
      detachInterrupt(digitalPinToInterrupt(MOTOR));  
      //
      // Im Bulb-Mode der M3 wird der Shutter viel später aufgezogen.
      // Deshalb anderes Timing vorgeben, wenn zuvor ein Motor-Interrupt kam.
      // Das wird daran erkannt, dass die Ausführung auf einer Warte-Anweisung
      //  steht
      //
      if (actionPtr == M_WAIT) {   // warten nach Motor-Interrupt?
        actionPtr = coilStopBulbAction;            //  weiter für Bulb-Abschluss
        //PRINTLN("c1_H_StopBulb");
      }
      else 
#endif
      {
        actionPtr = coilStopAction;
        //PRINTLN("c1_H_Stop");
      }
    }
#ifndef EOS_M100
    else if (triggerEvent == evMotorOn) {        // Motor-Start (vor c1_H oder nach c1_H):
# ifdef EOS_M3
      if (actionPtr == CS_WAIT) {                // zuletzt Motor-Interrupt in ...?
        //(Alternativ könnte hier auch auf (actionPtr != M_WAIT) geprüft werden!)
        actionPtr = coilFinishBulbAction;            //  weiter für Bulb-Abschluss
        //PRINTLN("Motor_H_StopBulb");
      }
      else 
# endif      
      {
        //PRINTLN("Motor_H");
        actionPtr = motorAction;
      }
    }
# if defined(EOS_M5) || defined(EOS_M50)
    else if (triggerEvent == evCoil2Start) {  // Start-Flanke Coil2:
      actionPtr = coil2StartAction;
      //PRINTLN("c2_L");
    }
    else if (triggerEvent == evCoil2Stop) {   // Stopp-Flanke (Coil2 inaktiv):
      actionPtr = coil2StopAction;
      //PRINTLN("c2_H_Stop");
    }
# endif      
#endif      
    triggerEvent = evNone;
  }
  //
  // Wenn das Ende einer Sequenz erreicht wurde, den Prozessor schlafen legen,
  //  andernfalls in der Loop() weitermachen:
  //
  if (actionPtr == 0) {
#ifdef ALLOW_SLEEP
    //set_sleep_mode(SLEEP_MODE_PWR_DOWN); // den tiefsten Schlaf auswählen PWR_DOWN
    set_sleep_mode(SLEEP_MODE_STANDBY); // nach 6 Takten wieder da, 35 µA
    sleep_enable();  // sleep mode einschalten
    sleep_mode();    // in den sleep mode gehen
    sleep_disable(); // sleep mode ausschalten nach dem Erwachen
#endif
  }
  else {                                            // noch zu tun:
    if (actionPtr->Time() <= DebugSlowdown(micros() - timingZero)) {  // jetzt?
      actionPtr->ApplyVal();     // Zustand [actionPtr] ausgeben.
      actionPtr++;                                  // actionPtr weiter,
#if defined(EOS_M5) || defined(EOS_M50)             // M5, M50: 
      if (actionPtr->Val() == MOTOR_VAL) {          //  jetzt auf den Motor warten?
        if (digitalRead(MOTOR) == HIGH) {           //  weiter, wenn Motor anläuft (nach 340ms im Bulb, 16ms sonst)
          actionPtr = motorAction;
          //PRINTLN("Motor_Start");     // (zeigt Mist, wenn lange gewartet (Bulb))
        }
        else {
          actionPtr--;                              // weiter warten, wenn Motor noch nicht an
          //PRINT(".");
        }
      }
#elif !defined(EOS_M100)
      if (actionPtr->Val() == MOTOR_VAL) {
        //motorSuppress = 1; // s.u.
        EIFR = 1 << digitalPinToInterrupt(MOTOR);  // clear flag for pending interrupt, http://gammon.com.au/interrupts
        attachInterrupt(digitalPinToInterrupt(MOTOR), motor_interrupt, MOTOR_TRIGGER);
        actionPtr++;                                // actionPtr weiter,
      }
#endif
      if (actionPtr->Val() == BLINK_VAL) {
        ToggleLED();
        actionPtr++;                              // actionPtr weiter,
      }
      if (actionPtr->Val() == STOP_VAL) {         //  löschen wenn fertig
        actionPtr = 0;
      }
    }
  }
  UpdateIntervallTimer();
}

inline bool IsActive(int coil)  {return digitalRead(coil) == COIL_ACTIVE;}
//////////////////////////////////////////////////////////////////////////////////////
// Interrupt-Handler
//
// a) Coil (Coil1)
//
void coil1_interrupt() {
  if (IsActive(COIL1)) {             // Coil1 wird aktiv:
    triggerEvent = evCoil1Start;     //  Flag für aktive Flanke
  }
  else {                            ////////////// Coil1 wird inaktiv:
    triggerEvent = evCoil1Stop;      // Ende
  }
  ToggleDebugLED();
}

// b) Coil2 (M50)
//
#ifdef COIL2
void coil2_interrupt() 
{
  if (IsActive(COIL2)) {            ////////////// Coil2 wird aktiv:
    triggerEvent = evCoil2Start;    //  Flag für aktive Flanke
  }
  else {                            ////////////// Coil2 wird inaktiv:
    triggerEvent = evCoil2Stop;     // Coil2 Ende
  }
  ToggleDebugLED();
}
#endif

// c) Motor: Triggert auf RISING. 
//  Hier unbesehen die Abschlussaktion anstoßen. Falls gerade ausgelöst wird, folgt
//   noch Coil. Dann wird diese Motor-Sequenz einfach überschrieben.   
// Alternative zur Erkennung der Auslösung: Motor-Sequenz nur nach evCoil1Stop
//  einleiten: if (previousEvent == evCoil1Stop)...
//  Beim Einschalten der Stromversorgung wird allerdings die Coil kurz aktiviert! Das müsste
//   dann wieder in Setup() init. werden: Auslasszähler = 1
//
#if !defined(EOS_M100) && !(defined(EOS_M50) || defined(EOS_M5))
void motor_interrupt() {
#ifdef MOTOR_GATE
  s.o.;  Eleganter wäre, den Interrupt temporär mit detachInterrupt(digitalPinToInterrupt(MOTOR)) stillzulegen
#endif  
#ifdef EOS_M3
  /*
  if (motorSuppress != 0) {
    // Seltsamerweise kommt bei der M3-Emulation sofort nach attachInterrupt ein
    // Interrupt, der unterdrückt werden muss!
    motorSuppress--;
    return;
  }
  */
  detachInterrupt(digitalPinToInterrupt(MOTOR));  // einmal reicht (Motor ist gepulst!)
#endif
  triggerEvent = evMotorOn;     // Flag für Motor-An
  ToggleDebugLED();
}
#endif


#endif // EOS_MXXX
