/////////////////////////////////////////////////////////////////////////////////////////////
//	ProtocolStack für Arduino Intervall-Timer mit Bluetooth-Modul
// 
//       18.10.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef protocol_h_
#define protocol_h_


// Comment out to force protocol into internal RAM:
//*
#if defined(SPIFLASH_CS)	// external Flash available:
#define PROTOCOL_FLASH		// (saves 360 Bytes of program storage space and 506 bytes of dynamic memory!)
#endif
//*/

#if defined(PROTOCOL_FLASH)
#include "FlashUpdate.h"
# define PROTOCOL_STACK_SIZE   0xFF00
#else
# define PROTOCOL_STACK_SIZE   500     // reduce if Arduino gets out of SDRam
  /*
  Jeder Eintrag in ProtocolStack enthält den Frameindex der Aufnahme seit dem Einschalten
  und für jeden Sensor einen Temperturwert in der Reihenfolge T_SENSOR_AVAILABLE, AHT20_AVAILABLE.
  Jeder der Einträge belegt ein int16_t, es sei denn, er wird komprimiert.
  (Protokoll im ext. Flash wird nicht komprimiert.)

  Sobald mindestens ein Wert komprimiert wird, wird ein komprimiertes Bytes (K-Byte)
  gespeichert, das immer daran zu erkennen ist, dass sein MSB gesetzt ist.

  - Wenn der Frameindex gleich dem vorherigen plus eins ist, entfällt sein Eintrag
    und das erste Bytes ist ein K-Byte. 
    Andernfalls ist das erste Wort der Frameindex (max. 0x7FFF). 
    Wenn diesem kein K-Byte folgt, ist es der erste, unkomprimierte Temperturwert.
    Wenn dann nochmals kein K-Byte folgt, ist das der zweite unkomprimierte Temperturwert.
  
  - Wenn sich ein Temperturwert um maximal +/-d von seinem Vorgänger unterscheidet,
    wird die Differenz in n Bits des K-Bytes geschrieben:
     Für nur einen Temperturwert sind das Bit 6-0 und d = 63.
     Für zwei Temperturwerte erhält T1 Bit 6-3 mit d = 7 und T2 erhält Bit 2-0 mit d = 3.
    Der Wert -d-1 zeigt jeweils an, dass ein voller Eintrag notwendig war, der dann
    dem K-Byte folgt. Zur einfacheren und  vorzeichenlosen Verarbeitung wird zur Differenz
    d+1 addiert, so dass dann 0 statt (-d-1) einen Überlauf der Differenz anzeigt.
  */

  #if defined(PROTOCOL_2T)
  const int dT1_shift = 3;
  const int dT1_limit = 8;
  const int dT2_shift = 0;
  const int dT2_limit = 4;
  const int dT2_mask  = (dT2_limit * 2 -1);
  #else
  const int dT1_shift = 0;
  const int dT1_limit = 64;
  #endif
  const int dT1_mask  = (dT1_limit * 2 -1);
#endif

class ProtocolStack {
  public:
	    ProtocolStack()   {Reset();}
    void    Reset();
    void    Setup();
    void    Stop()	      {active = active & ~1;}
    void    Continue()	      {active |= 1;}
    bool    IsActive() const  {return active & 1;}
    bool    HasOldData() const{return active & 2;}
    bool    IntegrateOldData();
    void    RemoveOldData();
    int	    GetCount() const  {return cnt;}
    bool    IsFull() const    {return nextWritePos > pbf + maxEntryOfs;}

    bool    IsSending() const {return sending != 0;}
    bool    IsSendingPaused() const	{return sending < 0;}
    bool    StartSending();
    void    ContinueSending()	{sending = 1;}
    void    PauseSending()	{sending = -1;}
    void    StopSending()	{sending = 0;}
    bool    ReadNextLine(int breakCount);

    #if defined(PROTOCOL_2T)
    bool    Add(int frame, int T1, int T2);
    int	    GetRead_T2() const  {return readT2;}
    enum { maxEntrySize = 8 }; // max Bytes (Index + Ts + Tu + timestamp)
    #else
    bool    Add(int frame, int T1);
    enum    { maxEntrySize = 4 };
    #endif
    int	    GetRead_I() const   {return readIndex;}
    int	    GetRead_T1() const  {return readT1;}
    #if defined(PROTOCOL_FLASH)
    int	    GetRead_tm() const  {return readtm;}
    #endif

    enum    { maxEntryOfs = PROTOCOL_STACK_SIZE - maxEntrySize };  // noch Platz für einen vollen Eintrag

  protected:
    uint16_t ReadWord();
    void    WriteWord(int w);
    #if !defined(PROTOCOL_FLASH)
    uint8_t ReadByte();
    int	    AppendReadByte(int8_t k_byte);
    void    WriteByte(int8_t b);
    int	    ExtractT1(int8_t k_byte);
    int	    ExtractT2(int8_t k_byte);
    void    DecompressT1(int8_t k_byte);
    void    DecompressT2(int8_t k_byte);
    #endif

  protected:
    #if defined(PROTOCOL_FLASH)
    uint16_t nextWritePos;
    uint16_t nextReadPos;
    uint32_t flashAddress(uint16_t rwa) const	  {return PROTOCOL_FLASHPAGE + uint16_t(rwa);}
    enum   {pbf = 0};			// relative Position in der flash page
    uint16_t readtm;
    #else
    int8_t  pbf[PROTOCOL_STACK_SIZE];
    int	    prevIndex;	// previuos index
    int	    prevT1;	// previuos T1
    #if defined(PROTOCOL_2T)
    int	    prevT2;	// previuos T2
    #endif
    int8_t* nextWritePos;
    int8_t* nextReadPos;
    #endif
    int	    cnt;	// index of next entry to add
    int	    readIndex;	// gelesener Index
    int	    readT1;	// gelesene T1
    #if defined(PROTOCOL_2T)
    int	    readT2;	// gelesene T1
    #endif
    int8_t  active;	// 0 wenn abgeschaltet, -1 wenn Flash noch nicht überschrieben
    int8_t  sending;
};

  
#endif
