/////////////////////////////////////////////////////////////////////////////////////////////
//	ProtocolStack für Arduino Intervall-Timer mit Bluetooth-Modul
// 
//       18.10.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////

#include "protocol.h"
//
// Inhalt löschen, Schreibposition auf 0, aktiv setzen.
//
//inline
void ProtocolStack::Reset()
{
  sending = 0;
  nextWritePos = pbf;	  // Schreibposition absolut / in der Flash Page
  cnt = 0;		  // nichts da
  active = 1;		  // aufnahmebereit

  #if !defined(PROTOCOL_FLASH) // nicht im Flash:
   prevT1 = 0;		  // noch kein T1,T2 geschrieben
   # if defined(PROTOCOL_2T)
   prevT2 = 0;
   # endif
  #endif
}


//
// Setup nach Start des Flash-Treibers
// 
//inline
void ProtocolStack::Setup()
{
  #if defined(PROTOCOL_FLASH) // im Flash: Erst jetzt Zugriff möglich!):
   //
   // Wenn altes Protokoll da ist, wird ein Flag gesetzt, so dass 
   //  es beim ersten Schreiben gelöscht wird. 
   // cnt wird gleich bestimmt, so dass eine Anzeige möglich ist.
   //  
   //
   nextReadPos = pbf;
   while (ReadWord() != 0xffff  &&  nextReadPos != pbf);    // Worte im Flash zählen 
   cnt = (nextReadPos - pbf - 2) / maxEntrySize;
   if (cnt > 0) {
     active = 2+1;	  // Flag: altes Protokoll noch vorhanden
     nextWritePos = pbf + maxEntrySize * cnt;	  // damit Anzeige klappt
   }
  #endif
}


//
// Alte Daten im Flash beibehalten.
// 
// Return false wenn nix da.
//
#if defined(PROTOCOL_FLASH) // im Flash:
//inline
bool ProtocolStack::IntegrateOldData()
{
  if (HasOldData()) {
    active &= 1;	  // nicht mehr alt
    return true;
  }
  return false;
}


//
// Altes Protokoll entfernen, wenn noch da und nicht integriert.
//
//inline
void ProtocolStack::RemoveOldData()
{
  if (HasOldData()) {
    cnt = 0;
    nextWritePos = 0;
    active &= 1;	  // nicht mehr alt
  }
}
#endif

//
// Ausgabe init.
// Return false wenn nix da.
//
//inline
bool ProtocolStack::StartSending()
{
  sending = GetCount() > 0 ? 1 : 0;
  nextReadPos = pbf;	  // Leseposition absolut
  #if !defined(PROTOCOL_FLASH)
  readIndex = 0;	// zuvor gelesener Index 
  readT1 = 0;		// zuvor gelesene T1
  # if defined(PROTOCOL_2T)
  readT2 = 0;		// zuvor gelesene T2
  # endif
  #endif
  return IsSending();
}


#if defined(PROTOCOL_FLASH)
//
// 16-Bit Wert eintragen, Schreibadresse weiter
//
//inline
void ProtocolStack::WriteWord(int w)
{
  if ((nextWritePos & 0x0FFF) == 0) {	// neuer 4K-Block?
    bootFlash.eraseSector(flashAddress(nextWritePos));  // sector erase time is 400ms
  }
  bootFlash.writeWord(flashAddress(nextWritePos), w);
  nextWritePos += 2;
}


//
// 16-Bit Wert entnehmen, Leseadresse weiter.
//
//inline
uint16_t ProtocolStack::ReadWord()
{
  uint16_t r = bootFlash.readWord(flashAddress(nextReadPos));
  nextReadPos += 2;
  return r;
}

#else	// ohne Flash, Komprimierung

//
// Byte entnehmen, Leseadresse weiter
//
//inline
uint8_t ProtocolStack::ReadByte()
{
  return *nextReadPos++;
}

//
// k_byte und das nächste Byte zu einem signed 15-Bit Wert kombinieren
//  und auf 16 Bit erweitern. 
// Leseadresse weiter.
//
//inline
int ProtocolStack::AppendReadByte(int8_t k_byte)
{
  return (int16_t(uint16_t(k_byte) << 9) >> 1) | ReadByte();
}

//
// Signed 15-Bit Wert entnehmen, auf 16 Bit erweitern.
// Leseadresse weiter.
//
//inline
uint16_t ProtocolStack::ReadWord()
{
  uint8_t r = ReadByte();
  return AppendReadByte(r);
}

//
// Byte eintragen, Schreibadresse weiter
//
//inline
void ProtocolStack::WriteByte(int8_t b)
{
  *nextWritePos++ = b;
}

//
// 16-Bit Wert eintragen, Schreibadresse weiter
//
//inline
void ProtocolStack::WriteWord(int w)
{
  WriteByte(uint8_t(w >> 8) & 0x7F); // Index komplett schreiben, HiByte zuerst,
  WriteByte(uint8_t(w));	     //  damit das höchste/erste Bit sicher 0 bleibt
}


//
// T1 aus K-Byte entnehmen
//
//inline
int ProtocolStack::ExtractT1(int8_t k_byte)
{
  return (k_byte >> dT1_shift) & dT1_mask;
}


#if defined(PROTOCOL_2T)
//
// T2 aus K-Byte entnehmen
//
//inline
int ProtocolStack::ExtractT2(int8_t k_byte)
{
  return (k_byte >> dT2_shift) & dT2_mask;
}
#endif

//
// T1 aus K-Byte entnehmen oder unkomprimiert lesen 
//  und in readT1 eintragen.
//
//inline
void ProtocolStack::DecompressT1(int8_t k_byte)
{
  int dT1 = ExtractT1(k_byte);
  if (dT1 == 0) {				// T1 ist nicht komprimiert:
    readT1 = ReadWord(); // lesen
  }
  else {
    readT1 += dT1 - dT1_limit;
  }
}

//
// T2 aus K-Byte entnehmen oder unkomprimiert lesen
//  und in readT2 eintragen.
//
//inline
void ProtocolStack::DecompressT2(int8_t k_byte)
{
  #if defined(PROTOCOL_2T)
  int dT2 = ExtractT2(k_byte);
  if (dT2 == 0) {				// T2 ist nicht komprimiert:
    readT2 = ReadWord(); // lesen
  }
  else {
    readT2 += dT2 - dT2_limit;
  }
  #else
  k_byte = k_byte;
  #endif
}
#endif

////////////////////////////////////////////////////////////////////////
// Nächste Zeile des Protokolls lesen, Position weitersetzen
//  und 'sending' ggf. abschalten
// Nach jeweils breakCount-1 Aufrufen einmal pausieren.
// 'sending' muss > 0 sein!
// 
// Rückgabe: true, wenn Zeile gelesen
//
//inline
bool ProtocolStack::ReadNextLine(int breakCount)
{
  if (nextReadPos < nextWritePos) {   // noch was da?
    if (sending > breakCount) {
      sending = 1;
      return false;
    }
    sending++;
    if (sending < 0) {
      sending = 1;
    }
#if defined(PROTOCOL_FLASH)	// ohne Zwischenpuffer müsste ReadNextLine() nach Ausgabe aufgerufen werden!
    readIndex = ReadWord();
    readT1 = ReadWord();
    #if defined(PROTOCOL_2T)
    readT2 = ReadWord();
    #endif
    readtm = ReadWord();
#else
    int8_t k_byte = ReadByte();
    readIndex++;				// Frameindex 1 weiter
    if (k_byte >= 0) {			        // Frameindex ist nicht komprimiert (ganz am Anfang oder nach Lücke):
      readIndex = (k_byte << 8) | ReadByte();	// lesen
      k_byte = ReadByte();			// nächstes Byte prüfen
      if (k_byte >= 0) {		        // T1 ist nicht komprimiert:
	readT1 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	#if defined(PROTOCOL_2T)
	k_byte = ReadByte();			// nächstes Byte prüfen
	if (k_byte >= 0) {			// T2 ist nicht komprimiert:
	  readT2 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	}
	else {
	  readT2 += ExtractT2(k_byte) - dT2_limit;
	}
	#endif
      }
      else {					// T1 ist komprimiert:
	readT1 += ExtractT1(k_byte) - dT1_limit;
	DecompressT2(k_byte);			// T2 ist evtl. komprimiert
      }
    }
    else {	        		        // Frameindex ist komprimiert:
      DecompressT1(k_byte);			// T1 ist evtl. komprimiert
      DecompressT2(k_byte);			// T2 ist evtl. komprimiert
    }
#endif
  }
  else {
    sending = 0;
  }
  return IsSending();
}


//
// Wert(e) in Protokoll eintragen
//
//inline
bool ProtocolStack::Add(int frame, int T1
    #if defined(PROTOCOL_2T)
    , int T2
    #endif
  ) 
{
  if (!IsActive() || IsFull()) {
    return false;
  }
#if defined(PROTOCOL_FLASH)		// ohne Zwischenpuffer müsste ReadNextLine() nach Ausgabe aufgerufen werden!
  RemoveOldData();			// ggf. altes Protokoll entfernen
  WriteWord(frame);
  WriteWord(T1);
  #if defined(PROTOCOL_2T)
    WriteWord(T2);
  #endif
  WriteWord(millis() / 1000);
#else
  uint8_t k_byte = 0x80;		// K-Byte erstellen:
  int dT1 = T1 - prevT1;		// T1-Diff + 8 = 1...15, wenn in Bit 6-3 unterzubringen
  if (abs(dT1) < dT1_limit) {		// passt:
    k_byte |= ((dT1 + dT1_limit) << dT1_shift);//  ins K-Byte eintragen
  }
  prevT1 = T1;
  #if defined(PROTOCOL_2T)
  int dT2 = T2 - prevT2;		// T2-Diff + 4 = 1...7, wenn in Bit 2-0 unterzubringen
  if (abs(dT2) < dT2_limit) {		// passt:
    k_byte |= ((dT2 + dT2_limit) << dT2_shift);//  ins K-Byte eintragen
  }
  prevT2 = T2;
  #endif
  prevIndex++;				// Folgeindex
  if (frame != prevIndex) {		// nicht letzter Index + 1:
    prevIndex = frame;
    WriteWord(frame);			//  Index komplett schreiben
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
      #if defined(PROTOCOL_2T)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      else {				// sonst jetzt das K-Byte
	WriteByte(k_byte);	// K-Byte steht in der Mitte
      }
      #endif
    }
    else {				// sonst jetzt das K-Byte
      WriteByte(k_byte);		// K-Byte steht in der Mitte
      #if defined(PROTOCOL_2T)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      #endif
    }
  }
  else {				// Index ist 1 weiter:
    WriteByte(k_byte);			// K-Byte steht vorn
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
    }
    #if defined(PROTOCOL_2T)
    if (abs(dT2) >= dT2_limit) {	// T2 passt nicht:
      WriteWord(T2);			//  komplett schreiben
    }
    #endif
  }
#endif
  cnt++;
  return true;
}  
